from toy_language import state, rstate, keywords, lexemes
import argparse
import pprint
import sys

class Tokenizer:
    def __init__(self):
        self.state = 'st_0'
        self.column = 1
        self.line   = 1
        self.pos = (1,1)
        self.inp = []

def st_0(tok, ch):
    if ch.isdigit():
        tok.pos = (tok.column, tok.line)
        tok.inp.append(ch)
        tok.state = 'st_digits'
    elif ch.isalpha() or ch == "_":
        tok.pos = (tok.column, tok.line)
        tok.inp.append(ch)
        tok.state = 'st_word'
    elif ch == " " or ch == "\n" or ch == "\t" or ch == "\r" or ch is "":
        pass
    elif ch == '"':
        tok.state = 'st_string'
    elif ch == "#":
        tok.state = 'st_comment'
    elif ch in keywords:
        advance('"' + ch + '"', ch,
            (tok.column, tok.line), (tok.column+1, tok.line))
    else:
        advance('OP', ch,
            (tok.column, tok.line), (tok.column+1, tok.line))

def st_comment(tok, ch):
    if ch == "\n":
        tok.state = 'st_0'

def st_word(tok, ch):
    if ch.isalpha() or ch == "_" or ch.isdigit():
        tok.inp.append(ch)
    else:
        text = "".join(tok.inp)
        if text in keywords:
            advance('"' + text + '"', text,
                tok.pos, (tok.column+1, tok.line))
        else:
            advance('IDENTIFIER', text, tok.pos, (tok.column+1, tok.line))
        tok.inp = []
        tok.state = 'st_0'
        st_0(tok, ch)

def st_digits(tok, ch):
    if ch.isdigit():
        tok.inp.append(ch)
    else:
        advance('NUMBER', "".join(tok.inp),
            tok.pos, (tok.column+1, tok.line))
        tok.inp = []
        tok.state = 'st_0'
        st_0(tok, ch)

def st_string(tok, ch):
    if ch == '"':
        advance('STRING', "".join(tok.inp),
            tok.pos, (tok.column,tok.line))
        tok.inp = []
        tok.state = 'st_0'
    elif ch == '\n':
        advance('STRING', "".join(tok.inp),
            tok.pos, (tok.column,tok.line))
        tok.inp = []
        tok.state = 'st_0'
    elif ch == '\\':
        tok.inp.append(ch)
        tok.state = st_string_esc
    else:
        tok.inp.append(ch)

def st_string_esc(tok, ch):
    tok.inp.append(ch)
    tok.state = st_string

# Collects every 'st_' into a dictionary.
tokenize_n = dict((k,v) for k,v in globals().items()
    if k.startswith('st_'))

def tokenize(tok, ch):
    tokenize_n[tok.state](tok, ch)
    if ch == "\n":
        tok.line += 1
        tok.column = 1
    else:
        tok.column += 1

# The parsing step
class Parser:
    def __init__(self):
        self.st = (None, 0, 0, 1) # (stack, top, layout, left)
        self.mode = 0
        self.syms   = []
        self.texts  = []
        self.starts   = []
        self.stops    = []
p = Parser()

def get_action(st, sym, left):
    dsym = sym if (st[2] < left or sym is None) else 'wfb'
    table = state[st[1]]
    if dsym in table:
        return table[dsym]
    else:
        return (8, 0)

# Parsing stack format: (st, obj)
def shift(st, sym, left, text):
    action, arg = get_action(st, sym, left)
    while action == 1:
        lhs, count = rstate[arg]
        objects = []
        for _ in range(count):
            st, obj = st[0]
            objects.append(obj)
        objects.reverse()
        robj = reduction(arg, objects)
        assert lhs is not None
        action, arg = state[st[1]][lhs]
        assert action != 1
        layout = st[2] if (action & 2 == 0) else st[3]
        st = (st, robj), arg, layout, st[3]
        action, arg = get_action(st, sym, left)
    if action & 8 != 0:
        return None
    if action & 4 != 0 and p.left != left:
        return None
    layout = st[2] if (action & 2 == 0) else left
    return (st, text), arg, layout, left

def finish(st):
    action, arg = get_action(st, None, 0)
    while action == 1:
        lhs, count = rstate[arg]
        objects = []
        for _ in range(count):
            st, obj = st[0]
            objects.append(obj)
        objects.reverse()
        if lhs is None:
            return True, objects[0]
        robj = reduction(arg, objects)
        action, arg = state[st[1]][lhs]
        assert action != 1
        st = (st, robj), arg, st[2], st[3]
        action, arg = get_action(st, None, 0)
    return False, None

def predict(st):
    for sym in state[st[1]]:
        if sym is None:
            pass
            #ok, ct = finish(st)
            #if ok:
            #    yield (sym
        elif sym in lexemes:
            ct = shift(st, sym, 1, None)
            if ct is not None:
                yield (sym, ct)

def predict_cont(st, depth):
    if depth == 0:
        for sym, ct in predict(st):
            yield [sym], ct
    else:
        for syms, ct in predict_cont(st, depth-1):
            for sym, dt in predict(ct):
                yield syms + [sym], dt

def advance(sym, text, start, stop):
    if p.mode == 0:
        st = shift(p.st, sym, start[0], text)
        if st is None:
            p.mode = 1
            p.syms.append(sym)
            p.texts.append(text)
            p.starts.append(start)
            p.stops.append(stop)
        else:
            p.st = st
    else:
        p.syms.append(sym)
        p.texts.append(text)
        p.starts.append(start)
        p.stops.append(stop)
        if len(p.syms) >= 20:
            repair()

def repair():
    cand = None
    for syms, ct in predict_cont(p.st, 3):
        for i, k in enumerate(editdistance(syms, p.syms)[len(syms)]):
            if cand is None:
                cand = k, i, syms
            elif k < cand[0] or (k == cand[0] and cand[1] >= i):
                cand = k, i, syms
    seq = rewritesequence(p.syms[:cand[1]], cand[2])
    syms = p.syms
    texts = p.texts
    starts = p.starts
    stops = p.stops
    stop = p.starts[0]
    p.mode = 0
    p.syms = []
    p.texts = []
    p.starts = []
    p.stops = []
    for act, arg in seq:
        if act in ('del', 'rw', 'shift'):
            sym = syms.pop(0)
            text = texts.pop(0)
            start = starts.pop(0)
            stop = stops.pop(0)
            if act == 'rw':
                sys.stderr.write("syntax error at line {}: rewrote {} to {} at col {}\n".format(
                    start[1], sym, arg, start[0]))
                sym = arg
                advance(sym, text, start, stop)
            elif act == 'shift':
                advance(sym, text, start, stop)
            elif act == 'del':
                sys.stderr.write("syntax error at line {}: deleted {} at col {}\n".format(
                    start[1], sym, start[0]))
        elif act == 'ins':
            sys.stderr.write("syntax error at line {}: inserted {} to col {}\n".format(
                stop[1], arg, stop[0]))
            advance(arg, None, stop, stop)
    while len(syms) > 0:
        sym = syms.pop(0)
        text = texts.pop(0)
        start = starts.pop(0)
        stop = stops.pop(0)
        advance(sym, text, start, stop)

def editdistance(s, t):
    m = len(s)
    n = len(t)
    d = [[i for j in range(n+1)] for i in range(m+1)]
    for j in range(1, n+1):
        d[0][j] = j
        for i in range(1, m+1):
            substitution_cost = int(s[i-1] != t[j-1])
            d[i][j] = min( 
                d[i-1][j] + 1,                        # deletion
                min(d[i][j-1] + 1,                    # insertion
                    d[i-1][j-1] + substitution_cost)) # substitution
    return d

def rewritesequence(s, t):
    d = editdistance(s, t)
    i = len(s)
    j = len(t)
    stack = []
    while i > 0 and j > 0:
        if d[i][j] == d[i][j-1]+1:
            stack.append(('ins', t[j-1]))
            j -= 1
        elif d[i][j] == d[i-1][j-1] + int(s[i-1] != t[j-1]):
            if s[i-1] == t[j-1]:
                stack.append(('shift', None))
            else:
                stack.append(('rw', t[j-1]))
            i -= 1
            j -= 1
        elif d[i][j] == d[i-1][j]+1:
            stack.append(('del', None))
            i -= 1
        else:
            assert False
    while j > 0:
        stack.append(('ins', t[j-1]))
        j -= 1
    while i > 0:
        stack.append(('del', None))
        i -= 1
    stack.reverse()
    return stack


def reduction(rule, args):
    if rule == 1:
        return []
    if rule == 2:
        return args[0] + [args[1]]
    if rule == 3:
        return [':block'] + args[1]
    return [rule] + args


parser = argparse.ArgumentParser(
    description='A test parser for toy language')
parser.add_argument('filename', type=str,
    help='The input file for the parser')
parser.add_argument('-p', '--printout',
    action='store_const', const=True, default=False,
    help='Print more details to stdout')
args = parser.parse_args()

tok = Tokenizer()
with open(args.filename, 'r') as fd:
    for ch in fd.read():
        tokenize(tok, ch)
tokenize(tok, "")
if p.mode != 0:
    repair()
result, tree = finish(p.st)
if result:
    if args.printout:
        pp = pprint.PrettyPrinter(indent=2)
        pp.pprint(tree)
else:
    print "parse error at eof"
